/*
 *  @(#) $Id: mdcom_kernel.c 899 2017-06-02 09:18:27Z ertl-honda $
 */
#include <linux/ioctl.h>
#include <stdbool.h>
#include <stdint.h>
#include <errno.h>
#include <assert.h>
#include "mdcom.h"
#include "mdcom_config.h"
#include "./driver/mdcom_driver.h"

/*
 * KERNEL INTERFACE
 */
static int fd;
pthread_mutex_t mutex;

int
__mdcom_kernel_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout, void *sregion)
{
	int ret;
	mdcom_init_data_t init_data;

	init_data.num_fifochs	= MDCOM_NUM_FIFOCHS;
	init_data.num_smemch	= MDCOM_NUM_SMEMCHS;
	init_data.sregion_pending_events = MDCOM_SREGION_PENDING_EVENTS;
	pthread_mutex_init(&mutex, NULL);

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_INIT, &init_data);
	if (ret == -1) {
		if(errno == EAGAIN) return MDCOM_EINIT;
		if(errno == EFAULT) return MDCOM_ESREGION;
		if(errno == EINVAL) return MDCOM_EPARAM;
		if(errno == ENOMEM) return MDCOM_ESREGION;
		assert(0);
	}

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_REGFIFOCH, fifoch_mask);
	if (ret == -1) {
		if(errno == ENOTTY) return MDCOM_NOINIT;
		if(errno == EALREADY) return MDCOM_EASSIGN;
		assert(0);
	}

	return 0;
}

int
__mdcom_kernel_term(void)
{
	close(fd);
	return 0;
}

/*
 * INIT Shared Region(Memory)
 */
int
__mdcom_kernel_sregion_init(void **sregion)
{
	errno = 0; /* clear last error */
	fd = open("/dev/mdcom", O_RDWR);
	if (fd < 0) {
		if(errno == EAGAIN) return MDCOM_EINIT;
		if(errno == EUSERS) return MDCOM_OPROCESS;
		assert(0);
	}

	errno = 0; /* clear last error */
	*sregion = mmap(NULL, MDCOM_SREGION_SIZE, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	if (*sregion == MAP_FAILED){
		if(errno == EAGAIN) return MDCOM_ESREGION;
		if(errno == EINVAL) return MDCOM_ESREGION;
		assert(0);
	}
	// Note: the shared memory is zero-ed by the RTOS at initialization

	return 0;
}

int
__mdcom_kernel_notify(mdcom_id_t fifoch_id)
{
	int ret;

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_NOTIFY, fifoch_id);
	if (ret == -1) {
		if(errno == ENOTTY) return MDCOM_NOINIT;
		if(errno == EINVAL) return MDCOM_EPARAM;
		if(errno == EACCES) return MDCOM_EASSIGN;
		assert(0);
	}

	return 0;
}

int
__mdcom_kernel_wait(mdcom_id_t fifoch_id, mdcom_time_t timeout)
{
	int ret;

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_WAIT, fifoch_id);
	if (ret == -1) {
		if(errno == ENOTTY) return MDCOM_NOINIT;
		if(errno == EINVAL) return MDCOM_EPARAM;
		if(errno == EACCES) return MDCOM_EASSIGN;
		if(errno == ERESTART) return MDCOM_RESTART;
		assert(0);
	}

	return 0;
}

int
__mdcom_kernel_get_fifoch_mask(mdcom_mask_t *fifoch_mask)
{
	int ret;

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_GETFIFOCHMASK, fifoch_mask);
	if (ret == -1) {
		if(errno == ENOTTY) return MDCOM_NOINIT;
		assert(0);
	}

	return 0;
}

int
__mdcom_kernel_fifoch_init(const mdcom_id_t fifoch_id)
{
	int ret;

	errno = 0; /* clear last error */
	ret = ioctl(fd, MDCOM_IOCTL_CMD_INITFIFO, fifoch_id);
	if (ret == -1) {
		if(errno == ENOTTY) return MDCOM_NOINIT;
		if(errno == EINVAL) return MDCOM_EPARAM;
		if(errno == EACCES) return MDCOM_EASSIGN;
		assert(0);
	}
	if (ret != 0 ) {
		assert(0);
	}

	return 0;
}

int
__mdcom_kernel_get_num_process(uint32_t *addr)
{
	return 0;
}


bool_t
__mdcom_kernel_mutex_trylock(uint32_t *p_lock)
{
#ifdef MDCOM_USE_DRIVER_LOCK
	bool_t locked = false;
	int ret;

	errno = 0; /* clear last error */
	do {
		ret = ioctl(fd, MDCOM_IOCTL_CMD_TRYLOCK, 0);
	}while( ret != 0 );
    
	if(*p_lock == 0) {
		*p_lock = 1;
		locked = true;
	}

	ioctl(fd, MDCOM_IOCTL_CMD_UNLOCK, 0);

	return locked;
#else /* !MDCOM_USE_DRIVER_LOCK */
	return __sync_bool_compare_and_swap (p_lock, 0, 1);
#endif /* MDCOM_USE_DRIVER_LOCK */
}

void
__mdcom_kernel_mutex_unlock(uint32_t *p_lock)
{
	*p_lock = 0;
}
