/*
 *  TOPPERS/MDCOM
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Multi Domain Communication Module
 *
 *  Copyright (C) 2012-2018 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: mdcom_user.c 983 2019-01-23 13:19:37Z ertl-honda $  
 */
#include "mdcom.h"
#include "mdcom_config.h"

#if (MDCOM_NUM_FIFOCHS == 0) && (MDCOM_NUM_SMEMCHS == 0)
#error "No channel defined. You should at least define 1 fifo channel or smem channel."
#endif /* (MDCOM_NUM_FIFOCHS == 0) && (MDCOM_NUM_SMEMCHS == 0) */

//#define MDCOM_DEBUG
#define DEBUG_OUT false

/*
 * INTERNAL TYPES
 */
typedef struct {
	uint32_t	*p_write;
	uint32_t	*p_read;
	uint32_t	*p_buffer;
	uint32_t	num_elements;
} mdcom_fifo_t;

typedef struct {
	bool_t			assigned;
	uint32_t		block_size;
	uint32_t		num_blocks;
	uint32_t		*p_blocks_buffer;
	mdcom_fifo_t	m2r_fifo;
	mdcom_fifo_t	r2m_fifo;
	mdcom_id_t		*p_m2r_filter_id;
	mdcom_id_t		*p_r2m_filter_id;
} mdcom_fifoch_t;

typedef struct {
	uint32_t	*p_lock;
	uint32_t	buffer_size;
	uint32_t	*p_buffer;
} mdcom_smemch_t;

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER

#define MDCOM_NUM_SRC2DEST_FILTERS MDCOM_NUM_M2R_FILTERS
#define MDCOM_NUM_DEST2SRC_FILTERS MDCOM_NUM_R2M_FILTERS

#define GET_SRC2DEST_FIFO(p_fifoch)				(p_fifoch->m2r_fifo)
#define GET_DEST2SRC_FIFO(p_fifoch)				(p_fifoch->r2m_fifo)
#define GET_SRC2DEST_FILTER_ID(p_fifoch)		(*(p_fifoch->p_m2r_filter_id))
#define SET_SRC2DEST_FILTER_ID(p_fifoch, id)	(*(p_fifoch->p_m2r_filter_id) = id)
#define GET_DEST2SRC_FILTER_ID(p_fifoch)		(*(p_fifoch->p_r2m_filter_id))
#define SET_DEST2SRC_FILTER_ID(p_fifoch, id)	(*(p_fifoch->p_r2m_filter_id) = id)

#define MDCOM_FIFOCH_DEFAULT_DEST2SRC_FILTERS MDCOM_FIFOCH_DEFAULT_R2M_FILTERS
#define MDCOM_FIFOCH_DEFAULT_SRC2DEST_FILTERS MDCOM_FIFOCH_DEFAULT_M2R_FILTERS

#else /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_REMOTE */

#define MDCOM_NUM_SRC2DEST_FILTERS MDCOM_NUM_R2M_FILTERS
#define MDCOM_NUM_DEST2SRC_FILTERS MDCOM_NUM_M2R_FILTERS

#define GET_SRC2DEST_FIFO(p_fifoch)				(p_fifoch->r2m_fifo)
#define GET_DEST2SRC_FIFO(p_fifoch)				(p_fifoch->m2r_fifo)
#define GET_SRC2DEST_FILTER_ID(p_fifoch)		(*(p_fifoch->p_r2m_filter_id))
#define SET_SRC2DEST_FILTER_ID(p_fifoch, id)	(*(p_fifoch->p_r2m_filter_id) = id)
#define GET_DEST2SRC_FILTER_ID(p_fifoch)		(*(p_fifoch->p_m2r_filter_id))
#define SET_DEST2SRC_FILTER_ID(p_fifoch, id)	(*(p_fifoch->p_m2r_filter_id) = id)

#define MDCOM_FIFOCH_DEFAULT_DEST2SRC_FILTERS MDCOM_FIFOCH_DEFAULT_M2R_FILTERS
#define MDCOM_FIFOCH_DEFAULT_SRC2DEST_FILTERS MDCOM_FIFOCH_DEFAULT_R2M_FILTERS

#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

typedef struct {
	bool_t				initialized;
	mdcom_fifoch_t		fifochs[MDCOM_NUM_FIFOCHS];
	mdcom_smemch_t		smemchs[MDCOM_NUM_SMEMCHS];
	mdcom_filter_t		filters[MDCOM_NUM_SRC2DEST_FILTERS];
} mdcom_user_t;

/*
 * MODULE VARIABLES
 */
static mdcom_user_t the_mdcom_user = {
	.initialized = false,
#ifdef MDCOM_FILTERS
	.filters     = MDCOM_FILTERS,
#endif /* MDCOM_FILTERS */
};

/*
 *  GET control block
 */
#define get_fifoch(fifoch_id)		(&(the_mdcom_user.fifochs[fifoch_id]))
#define get_smemch(smemch_id)		(&(the_mdcom_user.smemchs[smemch_id]))

/*
 * FIFO
 */
static inline int
__mdcom_fifo_empty(mdcom_fifo_t *p_fifo)
{
	return (*p_fifo->p_write == *p_fifo->p_read);
}

#ifdef COPY_UPDATE_ALGORITHM_ENABLE

static inline int
__mdcom_fifo_full(mdcom_fifo_t *p_fifo)
{
	uint32_t fifo_write;
	uint32_t fifo_read;

	fifo_write = *p_fifo->p_write;
	assert(fifo_write >= 0);
	assert(fifo_write < p_fifo->num_elements);

	fifo_read = *p_fifo->p_read;
	assert(fifo_read >= 0);
	assert(fifo_read < p_fifo->num_elements);

	return ((fifo_write + 1) % p_fifo->num_elements) == fifo_read;
}

static inline int
__mdcom_fifo_enqueue(mdcom_fifo_t *p_fifo, uint32_t value)
{
	uint32_t fifo_write;

	if (__mdcom_fifo_full(p_fifo)) return -1;

	fifo_write = *p_fifo->p_write;
	assert(fifo_write >= 0);
	assert(fifo_write < p_fifo->num_elements);

	p_fifo->p_buffer[fifo_write] = value;
	*p_fifo->p_write = (fifo_write + 1) % p_fifo->num_elements;

	return 0;
}

static inline int
__mdcom_fifo_dequeue(mdcom_fifo_t *p_fifo, uint32_t *p_value)
{
	uint32_t fifo_read;

	if (__mdcom_fifo_empty(p_fifo)) return -1;

	fifo_read = *p_fifo->p_read;
	assert(fifo_read >= 0);
	assert(fifo_read < p_fifo->num_elements);

	*p_value = p_fifo->p_buffer[*p_fifo->p_read];
	*p_fifo->p_read = (*p_fifo->p_read + 1) % p_fifo->num_elements;

	return 0;
}

#else /* !COPY_UPDATE_ALGORITHM_ENABLE */

static inline int
__mdcom_fifo_full(mdcom_fifo_t *p_fifo)
{
	return ((*p_fifo->p_write + 1) % p_fifo->num_elements) == *p_fifo->p_read;
}

static inline int
__mdcom_fifo_enqueue(mdcom_fifo_t *p_fifo, uint32_t value)
{
	if (__mdcom_fifo_full(p_fifo)) return -1;

	p_fifo->p_buffer[*p_fifo->p_write] = value;
	*p_fifo->p_write = (*p_fifo->p_write + 1) % p_fifo->num_elements;

	return 0;
}

static inline int
__mdcom_fifo_dequeue(mdcom_fifo_t *p_fifo, uint32_t *p_value)
{
	if (__mdcom_fifo_empty(p_fifo)) return -1;

	*p_value = p_fifo->p_buffer[*p_fifo->p_read];
	*p_fifo->p_read = (*p_fifo->p_read + 1) % p_fifo->num_elements;

	return 0;
}

#endif /* COPY_UPDATE_ALGORITHM_ENABLE */

extern void *memset(void *str, int c, size_t len);

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER

static inline void
__mdcom_fifo_init(mdcom_fifo_t *p_fifo, uint32_t num_blocks)
{
	*(p_fifo->p_read) = 0;
	*(p_fifo->p_write) = 0;
	memset(p_fifo->p_buffer, 0, sizeof(uint32_t)*(num_blocks + 1));
}

void
__mdcom_fifoch_init(mdcom_id_t fifoch_id)
{
	uint32_t *p_lock;
	uint32_t loop;
	mdcom_fifoch_t *p_fifoch;
	uint32_t dest2src_def_filter[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_DEFAULT_DEST2SRC_FILTERS;
	uint32_t src2dest_def_filter[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_DEFAULT_SRC2DEST_FILTERS;

	p_fifoch = get_fifoch(fifoch_id);

	__mdcom_fifo_init(&GET_SRC2DEST_FIFO(p_fifoch), p_fifoch->num_blocks);
	__mdcom_fifo_init(&GET_DEST2SRC_FIFO(p_fifoch), p_fifoch->num_blocks);

	for (loop=0; loop < p_fifoch->num_blocks; loop++){
		p_lock = p_fifoch->p_blocks_buffer + loop*(1 + (p_fifoch->block_size / sizeof(uint32_t)));
		*p_lock = 0;
	}

	// set default filter
	SET_DEST2SRC_FILTER_ID(p_fifoch, dest2src_def_filter[fifoch_id]);
	SET_SRC2DEST_FILTER_ID(p_fifoch, src2dest_def_filter[fifoch_id]);
}

#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

/*
 *  CHECK
 */
#define CHECK_NOINIT() do {						\
	if (the_mdcom_user.initialized == false) {	\
		return MDCOM_NOINIT;					\
	}											\
} while(false)

#define CHECK_FIFOCH_ID(id) do {	\
	if (id >= MDCOM_NUM_FIFOCHS) {	\
		return MDCOM_EPARAM;		\
	}								\
} while(false)

#define CHECK_SMEMCH_ID(id) do {	\
	if (id >= MDCOM_NUM_SMEMCHS) {	\
		return MDCOM_EPARAM;		\
	}								\
} while(false)

#define CHECK_BLOCK_ID(id, p_fifo) do {	\
	if (id >= p_fifo->num_blocks) {		\
		return MDCOM_EPARAM;			\
	}									\
} while(false)

#define CHECK_FILTER_ID(id) do {										\
	if ((id >= MDCOM_NUM_DEST2SRC_FILTERS) && (id != NULL_FILTER)		\
										&& (id != DEFAULT_FILTER)) {	\
		return MDCOM_EPARAM;											\
	}																	\
} while(false)

#if MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX
#define CHECK_ASSIGNED(p_fifoch)
#else  /* !(MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX) */
#define CHECK_ASSIGNED(p_fifoch) do {	\
	if (!(p_fifoch->assigned)) {		\
		return MDCOM_EASSIGN;			\
	}									\
} while(false)
#endif /* MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX */

#define CHECK_NULL(ptr) do {	\
	if (ptr == NULL) {			\
		return MDCOM_EPARAM;	\
	}							\
} while(false)

#define CHECK_CTX() do {					\
	if (!__mdcom_kernel_check_tskctx()) {	\
		return MDCOM_ECTX;					\
	}										\
} while(false)

#if MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX
#if MDCOM_NUM_FIFOCHS <= 32
#define CHECK_FIFOCHMASK(fifoch_mask) do {					\
	if ((fifoch_mask & (0xffffffff << MDCOM_NUM_FIFOCHS)) != 0U) {	\
		return MDCOM_EPARAM;								\
	}														\
} while(false)
#else  /* MDCOM_NUM_FIFOCHS */
#error Number of FIFO Channel must be 32 or less. 
#endif /* MDCOM_NUM_FIFOCHS */
#else  /* !(MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX) */
#define CHECK_FIFOCHMASK(fifoch_mask)
#endif /* MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX */


/*
 * DEBUG
 */
#ifdef MDCOM_DEBUG

#define MDCOM_DEBUG_INIT
#define MDCOM_DEBUG_EALLOC
#define MDCOM_DEBUG_GETBUFFER
#define MDCOM_DEBUG_ENQUEUE
#define MDCOM_DEBUG_DEQUEUE

static void
___mdcom_debug_print_locks(mdcom_fifoch_t *fifoch)
{
	int i;
	uint32_t *tmp;

	PRINT("LOCKS");
	for(i=0; i < fifoch->num_blocks; i++) {
		tmp = fifoch->p_blocks_buffer + i*(1 + (fifoch->block_size) / sizeof(uint32_t));
		PRINT("*(0x%p): 0x%x", tmp, *((uint32_t *)tmp));
	}
}

static void
___mdcom_debug_print_fifoch(mdcom_fifoch_t *fifoch)
{
	PRINT("block_size: %u bytes", fifoch->block_size);
	PRINT("num_blocks: %u", fifoch->num_blocks);
	PRINT("p_blocks_buffer: 0x%p", fifoch->p_blocks_buffer);
	PRINT("p_m2r_filter_id: %u", *fifoch->p_m2r_filter_id);
	PRINT("p_r2m_filter_id: %u", *fifoch->p_r2m_filter_id);
}

static void
___mdcom_debug_print_fifos(mdcom_fifoch_t *fifoch)
{
	int i;

	PRINT("m2r_fifo.p_write: %u", *fifoch->m2r_fifo.p_write);
	PRINT("m2r_fifo.p_read: %u", *fifoch->m2r_fifo.p_read);
	PRINT("m2r_fifo.p_buffer: 0x%p", fifoch->m2r_fifo.p_buffer);
	PRINT("m2r_fifo.num_elements: %u", (uint32_t)fifoch->m2r_fifo.num_elements);
	PRINT("m2r_fifo:");
	for(i=0; i < fifoch->num_blocks+1; i++) {
		PRINT("*(0x%p): 0x%x", fifoch->m2r_fifo.p_buffer + i, *(fifoch->m2r_fifo.p_buffer + i));
	}

	PRINT("r2m_fifo.p_write: %u", *fifoch->r2m_fifo.p_write);
	PRINT("r2m_fifo.p_read: %u", *fifoch->r2m_fifo.p_read);
	PRINT("r2m_fifo.p_buffer: 0x%p", fifoch->r2m_fifo.p_buffer);
	PRINT("r2m_fifo.num_elements: %u", (uint32_t)fifoch->r2m_fifo.num_elements);
	PRINT("r2m_fifo:");
	for(i=0; i < fifoch->num_blocks+1; i++) {
		PRINT("*(0x%p): 0x%x", fifoch->r2m_fifo.p_buffer + i, *(fifoch->r2m_fifo.p_buffer + i));
	}
}

static void
___mdcom_debug_print_blocks(mdcom_fifoch_t *fifoch)
{
	int i, j;
	uint32_t *tmp, *tmp2;

	PRINT("BLOCKS");
	for(i=0; i < fifoch->num_blocks; i++) {
		tmp = fifoch->p_blocks_buffer + i*(1 + fifoch->block_size) + 1;
		PRINT("block(%d): 0x%p:", i, tmp);
		for(j=0; j < fifoch->block_size; j++) {
			tmp2 = tmp + j;
			PRINT("*(0x%p): 0x%x", tmp2, *(tmp2));
		}
	}
}

static void
___mdcom_debug_print_all()
{
	int i;
	mdcom_fifoch_t *p_fifoch;

	PRINT("initialized: %s", (the_mdcom_user.initialized)?"true":"false");

	for(i=0; i<MDCOM_NUM_SRC2DEST_FILTERS; i++)
		PRINT("filter(%d): 0x%p", i, the_mdcom_user.filters[i]);

	for(i=0; i<MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch = &the_mdcom_user.fifochs[i];
		PRINT("CHANNEL %d", i);
		___mdcom_debug_print_fifoch(p_fifoch);
		___mdcom_debug_print_fifos(p_fifoch);
		___mdcom_debug_print_locks(p_fifoch);
		___mdcom_debug_print_blocks(p_fifoch);
	}
}
#endif /* MDCOM_DEBUG */

#ifdef MDCOM_DEBUG_ALLOC
#define __mdcom_debug_print_locks(fifoch)	___mdcom_debug_print_locks(fifoch)
#else /* !MDCOM_DEBUG_ALLOC */
#define __mdcom_debug_print_locks(fifoch)
#endif /* MDCOM_DEBUG_ALLOC */

#ifdef MDCOM_DEBUG_GETBUFFER
#define __mdcom_debug_print_fifoch(fifoch)	___mdcom_debug_print_fifoch(fifoch)
#else /* !MDCOM_DEBUG_GETBUFFER */
#define __mdcom_debug_print_fifoch(fifoch)
#endif /* MDCOM_DEBUG_GETBUFFER */

#ifdef MDCOM_DEBUG_ENQUEUE
#define __mdcom_debug_print_fifos(fifoch)	___mdcom_debug_print_fifos(fifoch)
#else  /* !MDCOM_DEBUG_ENQUEUE */
#define __mdcom_debug_print_fifos(fifoch)
#endif /* MDCOM_DEBUG_ENQUEUE */

#ifdef MDCOM_DEBUG_INIT
#define __mdcom_debug_print_all()	___mdcom_debug_print_all()
#else  /* !MDCOM_DEBUG_INIT */
#define __mdcom_debug_print_all()
#endif /* MDCOM_DEBUG_INIT */

/*
 * mdcom_init
 */
int
mdcom_init(mdcom_mask_t fifoch_mask, mdcom_time_t timeout)
{
	int ret, i;
	void *p_sregion;
	mdcom_fifoch_t *p_fifoch;
	mdcom_smemch_t *p_smemch;

	uint32_t num_blocks[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_NUM_BLOCKS;
	uint32_t block_size[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_BLOCK_SIZES;
	uint32_t fifoch_offsets[MDCOM_NUM_FIFOCHS] = MDCOM_SREGION_FIFOCH_OFFSETS;
	uint32_t buffer_size[MDCOM_NUM_SMEMCHS] = MDCOM_SMEMCH_BUFFER_SIZES;
	uint32_t smemch_offsets[MDCOM_NUM_SMEMCHS]  = MDCOM_SREGION_SMEMCH_OFFSETS;

	CHECK_CTX();
	CHECK_FIFOCHMASK(fifoch_mask);

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
	// 0. - check initilized 
	if (__mdcom_kernel_get_state() == MDCOM_KERNEL_INITD) {
		the_mdcom_user.initialized = true;
		return MDCOM_SUCCESS;
	}

	// 1.- create shared memory object and map it and 0 clean.
	if (__mdcom_kernel_get_state() == MDCOM_KERNEL_NOINIT) {
		ret = __mdcom_kernel_sregion_init(&p_sregion);
		if (ret != 0) return ret;
	}

	if (__mdcom_kernel_get_state() == MDCOM_KERNEL_REQINIT) {
		goto call__mdcom_kernel_init;
	}
#else /* MDCOM_TYPE_ROLL != MDCOM_TYPE_ROLL_MASTER */
	ret = __mdcom_kernel_sregion_init(&p_sregion);
	if (ret != 0) return ret;
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */

	// 2.- initialize all fifo channels
	for(i=0; i < MDCOM_NUM_FIFOCHS; i++) {
		p_fifoch = get_fifoch(i);

#if MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX
		//skip initialize in case of a fifo channel is not assignd
		if ((fifoch_mask & (1 << i)) == 0) {
			p_fifoch->assigned = false;
			continue;
		}
#endif /* MDCOM_TYPE_OS == MDCOM_TYPE_OS_LINUX */

		// local variables
		p_fifoch->assigned = true;
		p_fifoch->block_size = block_size[i];
		p_fifoch->num_blocks = num_blocks[i];
		p_fifoch->m2r_fifo.num_elements = p_fifoch->num_blocks + 1;
		p_fifoch->r2m_fifo.num_elements = p_fifoch->num_blocks + 1;
		// initialize pointers to shared memory
		p_fifoch->m2r_fifo.p_write  = p_sregion + fifoch_offsets[i];
		p_fifoch->m2r_fifo.p_read   = p_fifoch->m2r_fifo.p_write  + 1;
		p_fifoch->m2r_fifo.p_buffer = p_fifoch->m2r_fifo.p_read   + 1;
		p_fifoch->r2m_fifo.p_write  = p_fifoch->m2r_fifo.p_buffer + p_fifoch->m2r_fifo.num_elements;
		p_fifoch->r2m_fifo.p_read   = p_fifoch->r2m_fifo.p_write  + 1;
		p_fifoch->r2m_fifo.p_buffer = p_fifoch->r2m_fifo.p_read   + 1;
		p_fifoch->p_m2r_filter_id   = (mdcom_id_t*)(p_fifoch->r2m_fifo.p_buffer + p_fifoch->r2m_fifo.num_elements);
		p_fifoch->p_r2m_filter_id   = (mdcom_id_t*)(p_fifoch->p_m2r_filter_id   + 1);
		p_fifoch->p_blocks_buffer    = (uint32_t*)(p_fifoch->p_r2m_filter_id   + 1); 

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
		// initialize 
		__mdcom_fifoch_init(i);
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	}

	// 2.- initialize all smemchs
	for(i=0; i<MDCOM_NUM_SMEMCHS; i++) {
		p_smemch = get_smemch(i);
		p_smemch->buffer_size = buffer_size[i];
		p_smemch->p_lock = p_sregion + smemch_offsets[i];
		p_smemch->p_buffer = p_smemch->p_lock + 1;
	}

	__mdcom_debug_print_all();

#if MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER
  call__mdcom_kernel_init:
#endif /* MDCOM_TYPE_ROLL == MDCOM_TYPE_ROLL_MASTER */
	// 4.- init events and wait for the Destinataion OS to initialize (with a timeout)
	ret = __mdcom_kernel_init(fifoch_mask, timeout, p_sregion);
	if (ret != 0) return ret;

	the_mdcom_user.initialized = true;

	return MDCOM_SUCCESS;
}

/*
 * mdcom_term
 */
int
mdcom_term(void)
{
	CHECK_NOINIT();

	the_mdcom_user.initialized = false;
	return __mdcom_kernel_term();
}

/*
 * mdcom_sense_process
 */
int
mdcom_sense_process(uint32_t *p_num_process) {
	CHECK_NOINIT();

	return __mdcom_kernel_get_num_process(p_num_process);
}

/*
 * mdcom_sense_fifoch
 */
int
mdcom_sense_fifoch(mdcom_mask_t *p_fifoch_mask) {
	CHECK_NOINIT();

	return __mdcom_kernel_get_fifoch_mask(p_fifoch_mask);
}

/*
 * mdcom_fifoch_alloc
 */
int
mdcom_fifoch_alloc(const mdcom_id_t fifoch_id,
								mdcom_id_t *p_block_id)
{
	int i, ret;
	mdcom_fifoch_t *p_fifoch;
	uint32_t *p_lock;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_NULL(p_block_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);

	for(i=0; i < p_fifoch->num_blocks; i++) {
		p_lock = p_fifoch->p_blocks_buffer + i*(1 + (p_fifoch->block_size / sizeof(uint32_t)));
		// atomic: if (*p_lock == 0) *p_lock = 1, ret = true
		ret =  __mdcom_kernel_mutex_trylock(p_lock);
		if (ret == false) continue;
		*p_block_id = (mdcom_id_t)i;

		__mdcom_debug_print_locks(p_fifoch);

		return MDCOM_SUCCESS;
	}

	__mdcom_debug_print_locks(p_fifoch);

	return MDCOM_EALLOC;
}

/*
 * mdcom_fifoch_free
 */
int
mdcom_fifoch_free(const mdcom_id_t fifoch_id, const mdcom_id_t block_id)
{
	mdcom_fifoch_t *p_fifoch;
	uint32_t *p_lock;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);
	CHECK_BLOCK_ID(block_id, p_fifoch);

	p_lock = p_fifoch->p_blocks_buffer + block_id*(1 + (p_fifoch->block_size / sizeof(uint32_t)));
	if (*p_lock == 0) return MDCOM_EALLOC;

	__mdcom_kernel_mutex_unlock(p_lock);

	__mdcom_debug_print_locks(p_fifoch);

	return MDCOM_SUCCESS;
}

/*
 * mdcom_fifoch_get
 */
int
mdcom_fifoch_get(const mdcom_id_t fifoch_id, const mdcom_id_t block_id,
					void **p_block)
{
	mdcom_fifoch_t *p_fifoch;
	uint32_t *p_lock;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);
	CHECK_BLOCK_ID(block_id, p_fifoch);

	p_lock = p_fifoch->p_blocks_buffer + block_id*(1 + (p_fifoch->block_size/ sizeof(uint32_t)));
	if (*p_lock == 0) return MDCOM_EALLOC;

	*p_block = (void *)(p_lock + 1);

	__mdcom_debug_print_fifoch(p_fifoch);

	return MDCOM_SUCCESS;
}

/*
 * mdcom_fifoch_enqueue
 */
int
mdcom_fifoch_enqueue(const mdcom_id_t fifoch_id,
							const mdcom_id_t block_id)
{
	mdcom_fifoch_t *p_fifoch;
	uint32_t *p_lock;
	int ret;
	mdcom_id_t filter_id;
	mdcom_filter_t filter;
	void *buffer;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);
	CHECK_BLOCK_ID(block_id, p_fifoch);

	// filtering
	filter_id = GET_SRC2DEST_FILTER_ID(p_fifoch);
	if (filter_id != NULL_FILTER) {
		assert(filter_id < MDCOM_NUM_SRC2DEST_FILTERS);
		filter = the_mdcom_user.filters[filter_id];
		// TODO: check this address
		buffer = p_fifoch->p_blocks_buffer + block_id*(1 + (p_fifoch->block_size/sizeof(uint32_t))) + 1;
		if (filter(buffer, p_fifoch->block_size) == false) return MDCOM_FILTER;
	}
	p_lock = p_fifoch->p_blocks_buffer + block_id*(1 + (p_fifoch->block_size/ sizeof(uint32_t)));
	if (*p_lock == 0) return MDCOM_EALLOC;

	__mdcom_kernel_lock();
	ret = __mdcom_fifo_enqueue(&GET_SRC2DEST_FIFO(p_fifoch), block_id);
	__mdcom_kernel_unlock();
	if(ret == -1) return MDCOM_FULL;

	__mdcom_debug_print_fifos(p_fifoch);

	return 0;
}

/*
 * mdcom_fifoch_dequeue
 */
int
mdcom_fifoch_dequeue(const mdcom_id_t fifoch_id,
							mdcom_id_t *p_block_id)
{
	mdcom_fifoch_t *p_fifoch;
	int ret;
	uint32_t *p_lock;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_NULL(p_block_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);

	__mdcom_kernel_lock();
	ret = __mdcom_fifo_dequeue(&GET_DEST2SRC_FIFO(p_fifoch), (uint32_t*)p_block_id);
	__mdcom_kernel_unlock();
	if (ret == -1) return MDCOM_EMPTY;

	assert(*p_block_id < p_fifoch->num_blocks);

	p_lock = p_fifoch->p_blocks_buffer + *p_block_id*(1 + (p_fifoch->block_size / sizeof(uint32_t)));
	if (*p_lock == 0) return MDCOM_EALLOC;

	__mdcom_debug_print_fifos(p_fifoch);

	return 0;
}

/*
 * mdcom_fifoch_notify
 */
int
mdcom_fifoch_notify(const mdcom_id_t fifoch_id)
{
	int ret;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_ASSIGNED(get_fifoch(fifoch_id));
	ret = __mdcom_kernel_notify(fifoch_id);
	assert(ret == 0);

	return 0;
}

/*
 * mdcom_fifoch_wait
 */
int
mdcom_fifoch_wait(const mdcom_id_t fifoch_id,
						 const mdcom_time_t timeout)
{
	int ret;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_ASSIGNED(get_fifoch(fifoch_id));

	if (!__mdcom_kernel_check_tskctx()) return MDCOM_ECTX;

	ret = __mdcom_kernel_wait(fifoch_id, timeout);

	return ret;
}

/*
 * mdcom_filter_set
 */
int mdcom_filter_set(const mdcom_id_t fifoch_id,
								const mdcom_id_t filter_id)
{
	mdcom_fifoch_t *p_fifoch;
	uint32_t def_filter[MDCOM_NUM_FIFOCHS] = MDCOM_FIFOCH_DEFAULT_DEST2SRC_FILTERS;

	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_FILTER_ID(filter_id);
	p_fifoch = get_fifoch(fifoch_id);
	CHECK_ASSIGNED(p_fifoch);

	if (filter_id == DEFAULT_FILTER) {
		SET_DEST2SRC_FILTER_ID(p_fifoch, def_filter[fifoch_id]);
	}
	else {
		SET_DEST2SRC_FILTER_ID(p_fifoch, filter_id);
	}

	return MDCOM_SUCCESS;
}

/*
 * mdcom_fifoch_init
 */
int
mdcom_fifoch_init(mdcom_id_t fifoch_id) {
	CHECK_NOINIT();
	CHECK_FIFOCH_ID(fifoch_id);
	CHECK_ASSIGNED(get_fifoch(fifoch_id));

	__mdcom_kernel_fifoch_init(fifoch_id);

	return MDCOM_SUCCESS;
}

/*
 * mdcom_smemch_get
 */
int
mdcom_smemch_get(const mdcom_id_t smemch_id, void **p_smem)
{
	mdcom_smemch_t *p_smemch;

	CHECK_NOINIT();
	CHECK_SMEMCH_ID(smemch_id);
	p_smemch = get_smemch(smemch_id);

	*p_smem = (void *)(p_smemch->p_buffer);

	return MDCOM_SUCCESS;
}

/*
 * mdcom_smemch_trylock
 */
int
mdcom_smemch_trylock(const mdcom_id_t smemch_id)
{
	mdcom_smemch_t *p_smemch;
	uint32_t *p_lock;
	int ret;

	CHECK_NOINIT();
	CHECK_SMEMCH_ID(smemch_id);
	p_smemch = get_smemch(smemch_id);

	p_lock = p_smemch->p_lock;
	for(;;){
		if (*p_lock == 1) return MDCOM_LOCKED;
		ret =  __mdcom_kernel_mutex_trylock(p_lock);
		if (ret == true) break;
	}

	return MDCOM_SUCCESS;
}

/*
 * mdcom_smemch_unlock
 */
int
mdcom_smemch_unlock(const mdcom_id_t smemch_id)
{
	mdcom_smemch_t *p_smemch;
	uint32_t *p_lock;

	CHECK_NOINIT();
	CHECK_SMEMCH_ID(smemch_id);
	p_smemch = get_smemch(smemch_id);

	p_lock = p_smemch->p_lock;
	if (*p_lock == 0) return MDCOM_UNLOCKED;
	__mdcom_kernel_mutex_unlock(p_lock);

	return MDCOM_SUCCESS;
}
