/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 *
 *  Copyright (C) 2000-2003 by Embedded and Real-Time Systems Laboratory
 *                              Toyohashi Univ. of Technology, JAPAN
 *  Copyright (C) 2004-2013 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 *
 *  上記著作権者は，以下の(1)〜(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  $Id: sample1_m.c 972 2019-01-18 09:03:11Z ertl-honda $
 */

#include <kernel.h>
#include <t_syslog.h>
#include <t_stdlib.h>
#include "syssvc/serial.h"
#include "syssvc/syslog.h"
#include "syssvc/logtask.h"
#include "kernel_cfg.h"
#include "sample1.h"
#include "mdcom.h"
#include "mdcom_config.h"

/*
 *  サービスコールのエラーのログ出力
 */
Inline void
svc_perror(const char *file, int_t line, const char *expr, ER ercd)
{
	if (ercd < 0) {
		t_perror(LOG_ERROR, file, line, expr, ercd);
	}
}

#define SVC_PERROR(expr) svc_perror(__FILE__, __LINE__, #expr, (expr))

/*
 * Periodic task
 */
void task(intptr_t exinf)
{
	static int i = 0;

	syslog(LOG_EMERG, "task %d", i++);
}

#ifdef TOPPERS_SAFEG_SECURE

#include "syscalls_api.h"

void btask(intptr_t exinf)
{
	uint32_t ret;
	uint32_t nt_start_address;

	syslog(LOG_NOTICE, "BTASK for Core%d Starts", exinf);

	while((nt_start_address = target_get_nt_start_address()) == -1);

	syslog(LOG_NOTICE, "%d : SafeG First switch to NT : address 0x%x", exinf, nt_start_address);
	ret = safeg_syscall_setntpc(nt_start_address);

	if (ret != SAFEG_SYSCALL_ERROR__OK) {
		syslog(LOG_NOTICE, "%d : SafeG ERROR: setntpc %u", exinf, ret);
		return;
	}

	while(1) {
		SIL_PRE_LOC;

		SIL_LOC_INT();
		safeg_load_idle();
		ret = safeg_syscall_switch();
		SIL_UNL_INT();

		if (ret != SAFEG_SYSCALL_ERROR__OK) {
			syslog(LOG_NOTICE, "%d : SafeG ERROR: switch %u", exinf, ret);
			break;
		}
	}
}
#endif /* TOPPERS_SAFEG_SECURE */

#ifdef ENABLE_TASK
/*
 * Cyclic handler
 */
void cyclic_handler(intptr_t exinf)
{
	SVC_PERROR(iact_tsk(TASK1_1));
}
#endif /* ENABLE_TASK */

static void *memcpy(void *dest, const void *src, size_t count) {
	char *dst8 = (char *)dest;
	char *src8 = (char *)src;

	while (count--) *dst8++ = *src8++;
	return dest;
}

#define SHMEM_SIZE 64

void
share_mem_test() {
	int ret;
	void *buffer;
	uint32_t size;
	int data;
	int i;

	size = 0;
	
	syslog(LOG_NOTICE, "Share mem test start!.");

	ret = mdcom_smemch_get(MDCOM_SHMEM_1, &buffer);
	if(ret != 0) {
		syslog(LOG_NOTICE, "Get Buffer Error!! %d",ret);
	}

	syslog(LOG_NOTICE, "buffer 0x%x, Size %d.", buffer, size);

	data = 0;
	while(1) {
		for(i = 0; i < SHMEM_SIZE; i++){
			*((uint32_t *)buffer + i) = data + i;
		}
		syslog(LOG_NOTICE, "Write data %d to %d", data, data + SHMEM_SIZE);
		data++;
		tslp_tsk(4000);
	}
}

void
main_task_tex_routine(TEXPTN texptn, intptr_t exinf){
	act_tsk(MAIN_TASK1);
	ext_tsk();
}

void main_task(intptr_t exinf)
{
    ER_UINT ercd;
    int ret;
    mdcom_id_t block_id;
    void *buffer;
    uint32_t size;
    char msg[] = "hello world";

    size = 0;

    SVC_PERROR(syslog_msk_log(0, LOG_UPTO(LOG_DEBUG)));

    logtask_flush(0);
    syslog(LOG_NOTICE, "MDCOM basic test program starts (exinf = %d).", (int_t) exinf);

    ercd = serial_opn_por(TASK_PORTID_G_SYSLOG);
    if (ercd < 0 && MERCD(ercd) != E_OBJ) {
	syslog(LOG_ERROR, "%s (%d) reported by `serial_opn_por'.",
			   itron_strerror(ercd), SERCD(ercd));
    }
    SVC_PERROR(serial_ctl_por(TASK_PORTID_G_SYSLOG,
	(IOCTL_CRLF | IOCTL_FCSND | IOCTL_FCRCV)));

#ifdef ENABLE_TASK
    SVC_PERROR(sta_cyc(CYCHDR1));
#endif

#ifdef DISABLE_TIMER
    SVC_PERROR(dis_int(IRQNO_TIMER));
#endif

    ena_tex();

    syslog(LOG_NOTICE, "Master mdcom_init");

    ret = mdcom_init(0x3, -1);
    assert(ret == 0);

    syslog(LOG_NOTICE, "MDCOM communications initialized");

    /* phase 1 */
    syslog(LOG_NOTICE, "Master send message in channel 0");

    ret = mdcom_fifoch_alloc(MDCOM_CHANNEL_0, &block_id);
    assert(ret == 0);

    ret = mdcom_fifoch_get(MDCOM_CHANNEL_0, block_id, &buffer);
    assert(ret == 0);
    syslog(LOG_NOTICE, "buffer 0x%x, Size %d.", buffer, size);

    memcpy(buffer, (void *)msg, sizeof(msg));

    ret = mdcom_fifoch_enqueue(MDCOM_CHANNEL_0, block_id);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master send event");
    ret = mdcom_fifoch_notify(MDCOM_CHANNEL_0);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master wait for ack");
    ret = mdcom_fifoch_wait(MDCOM_CHANNEL_0, -1);
    assert(ret == 0);


    /* phase 2 */
    syslog(LOG_NOTICE, "Master send params 3 and 5");

    *((uint32_t *)buffer) = 3;
    *((uint32_t *)(buffer + 4)) = 5;

    ret = mdcom_fifoch_notify(MDCOM_CHANNEL_0);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master wait result");
    ret = mdcom_fifoch_wait(MDCOM_CHANNEL_0, -1);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master rpc result: 3 x 5 = %u", *((uint32_t *)buffer));


    /* phase 3 */
    syslog(LOG_NOTICE, "Master send params 4 and 6");

    *((uint32_t *)buffer) = 4;
    *((uint32_t *)(buffer + 4)) = 6;

    ret = mdcom_fifoch_notify(MDCOM_CHANNEL_0);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master wait result");
    ret = mdcom_fifoch_wait(MDCOM_CHANNEL_0, -1);
    assert(ret == 0);

    syslog(LOG_NOTICE, "Master rpc result: 4 x 6 = %u", *((uint32_t *)buffer));

    ret = mdcom_fifoch_notify(MDCOM_CHANNEL_0);
    assert(ret == 0);

    /* phase 4 */
    syslog(LOG_NOTICE, "Master wait for sentence from GPOS in channel 1");
    ret = mdcom_fifoch_wait(MDCOM_CHANNEL_1, -1);
    assert(ret == 0);

    ret = mdcom_fifoch_dequeue(MDCOM_CHANNEL_1, &block_id);
    assert(ret == 0);

    ret = mdcom_fifoch_get(MDCOM_CHANNEL_1, block_id, &buffer);
    assert(ret == 0);

    syslog(LOG_NOTICE, "MSG: %s", (char *)buffer);

    syslog(LOG_NOTICE, "Master mdcom_block_free");
    ret = mdcom_fifoch_free(MDCOM_CHANNEL_1, block_id);
    assert(ret == 0);


    /* phase 5 */
    syslog(LOG_NOTICE, "Master wait for EVEN messages");
    ret = mdcom_filter_set(MDCOM_CHANNEL_0, MDCOM_FILTER_ID_EVEN);
    assert(ret == 0);

    ret = mdcom_fifoch_notify(MDCOM_CHANNEL_0);
    assert(ret == 0);

    ret = mdcom_fifoch_wait(MDCOM_CHANNEL_0, -1);
    assert(ret == 0);

    while(1) {
        ret = mdcom_fifoch_dequeue(MDCOM_CHANNEL_0, &block_id);
        if (ret == MDCOM_EMPTY) break;
        assert(ret == 0);

        ret = mdcom_fifoch_get(MDCOM_CHANNEL_0, block_id, &buffer);
        assert(ret == 0);

        syslog(LOG_NOTICE, "number: %u", *((uint32_t *)buffer));

        ret = mdcom_fifoch_free(MDCOM_CHANNEL_0, block_id);
        assert(ret == 0);
    }

    syslog(LOG_NOTICE, "Master com done!!");

    share_mem_test();

    SVC_PERROR(slp_tsk());

    SVC_PERROR(ext_ker());
    assert(0);
}
