/*
 *  TOPPERS/MDCOM
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Multi Domain Communication Module
 *
 *  Copyright (C) 2016 by Embedded and Real-Time Systems Laboratory
 *     Graduate School of Information Science, Nagoya University, JAPAN
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: mdcom_target.h 976 2019-01-18 15:34:24Z ertl-honda $
 */
#include "sil.h"
#include "t_stddef.h"

#define MDCOM_TARGET_CA53

#ifdef TOPPERS_ZYNQMP_A53_32
#define MDCOM_TARGET_CA53_32
#define MDCOM_DOMAIN_NAME  "FMP(CA53-AArch32)"
#include "arm.h"
#elif defined(TOPPERS_ZYNQMP_A53_64)
#define MDCOM_TARGET_CA53_64
#define MDCOM_DOMAIN_NAME  "FMP(CA53-AArch64)"
#include "arm64.h"
#include "zynqmp_a53.h"
#else /* none */
#error TOPPERS_ZYNQMP_A53_32 or TOPPERS_ZYNQMP_A53_64
#endif /* TOPPERS_ZYNQMP_A53_32 */

#define MDCOM_TYPE_API  MDCOM_TYPE_API_TOPPERS
#define MDCOM_TYPE_OS   MDCOM_TYPE_OS_FMP
#define MDCOM_TYPE_EMV  MDCOM_TYPE_EMV_AMP
#define MDCOM_TYPE_ROLL MDCOM_TYPE_ROLL_REMOTE

#include "../mdcom_target_common.h"

#define MDCOM_INT_PRIORITY		-15
#define MDCOM_INTNO			(XIPI_INTNO)
#define MDCOM_INTHO			(MDCOM_INTNO)

#define SREGION_PHYS_ADDRESS MDCOM_SHMEM_ADDRESS

static inline void
__mdcom_dcache_flush(void)
{
#ifdef MDCOM_TARGET_CA53_32
	dcache_clean();
#endif /* MDCOM_TARGET_CA53_32 */
}

#if MDCOM_TYPE_API == MDCOM_TYPE_API_TOPPERS

static inline bool_t
__mdcom_kernel_mutex_trylock(uint32_t *p_lock)
{
	int locked;
	bool_t ret = false;
	SIL_PRE_LOC;

	SIL_LOC_SPN();

#ifdef MDCOM_TARGET_CA53_32
	Asm("   mov      r2, #0x01   \n"
		"\t ldrex    r1, [%1]    \n"
		"\t cmp      r1, #0x00   \n"
		"\t strexeq  r1, r2,[%1] \n"
		"\t mov      %0, r1      \n"
		:"=r"(locked):"r"(p_lock):"r1","r2", "cc");
	
	if(locked == 0) {
		/* ロック取得成功 */
		SIL_DMB;
		Asm("":::"memory");
		ret = true;
	}
#else /* MDCOM_TARGET_CA53_64 */
	Asm("   mov   w2, #0x01    \n"
		"\t ldxr  w1, [%1]     \n"
		"\t cmp   w1, #0x00    \n"
		"\t b.ne  1f           \n"
		"\t stxr  w1, w2, [%1] \n"
		"\t 1:                 \n"
		"\t mov   %0, x1       \n"
		:"=r"(locked):"r"(p_lock):"x1", "x2", "cc");

	if(locked == 0) {
		/* ロック取得成功 */
		DATA_MEMORY_BARRIER();
		Asm("":::"memory");
		ret = true;
	}
#endif /* MDCOM_TARGET_CA53_32 */

	SIL_UNL_SPN();
 
	return ret;
}


static inline void
__mdcom_kernel_mutex_unlock(uint32_t *p_lock)
{
	*p_lock = 0;

	Asm("":::"memory");

#ifdef  MDCOM_TARGET_CA53_32
	SIL_DMB;
	*p_lock = 0;
	SIL_DMB;
#else /* MDCOM_TARGET_CA53_64 */
	DATA_MEMORY_BARRIER();
	*p_lock = 0;
	DATA_MEMORY_BARRIER();
#endif  /* MDCOM_TARGET_CA53_32 */
}

#endif /* MDCOM_TYPE_API  MDCOM_TYPE_API_TOPPERS */

