/*
 * MDCOM Multi Domain Communications Inter-processor interrupt Driver for Zynq ZCU102 APU
 *
 * Copyright (C) Nagoya University
 * Daniel Sangorrin <daniel.sangorrin@gmail.com>
 * Shinya Honda     <honda@ertl.jp>  
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  @(#) $Id: mdcom-ipi.c 897 2017-05-24 07:30:19Z ertl-honda $
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/of_irq.h>
#include <linux/of_platform.h>
#include <linux/slab.h>
#include <linux/cpu.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include "mdcom_target_driver.h"
#ifdef USE_RPU_NOTIFY_POLLING
#include <linux/kthread.h>
#include <linux/delay.h>
#endif  /* USE_RPU_NOTIFY_POLLING */

/* IPI reg offsets */
#define TRIG_OFFSET         0x00000000
#define OBS_OFFSET          0x00000004
#define ISR_OFFSET          0x00000010
#define IMR_OFFSET          0x00000014
#define IER_OFFSET          0x00000018
#define IDR_OFFSET          0x0000001C
#define IPI_ALL_MASK        0x0F0F0301


/* IPI mask */
#define RPU_IPI_INIT_MASK   0x00000100
#define RPU_IPI_MASK(n)     (RPU_IPI_INIT_MASK << n)
#define RPU_0_IPI_MASK      RPU_IPI_MASK(0)
#define RPU_1_IPI_MASK      RPU_IPI_MASK(1)
#define APU_IPI_MASK        0x00000001


/*
 *  Global variables
 */
static void __iomem         *apu_ipi_base =  NULL;
static void __iomem         *rpu0_ipi_base = NULL;
#ifdef USE_RPU_NOTIFY_POLLING
static struct task_struct   *kthread = NULL;
static volatile int         kthread_run = false;
#else   /* USE_RPU_NOTIFY_POLLING */
static u32                  ipi_irq = 0xFFFFFFFF;
#endif  /* USE_RPU_NOTIFY_POLLING */


/*
 *  Macros
 */
/* Register access */
#define reg_read(base, reg) \
    readl(((void __iomem *)(base)) + (reg))
#define reg_write(base, reg, val) \
    writel((val), ((void __iomem *)(base)) + (reg))


/*
 *  Prototypes
 */
extern irqreturn_t  mdcom_irq_handler(int irq, void *dev_id);


#ifdef USE_RPU_NOTIFY_POLLING
/*
 *  mdcom_poll_rpu()
 *
 *  Function : This function clears event of interupt.
 *  Argument : None
 *  Return :   None
 */
static void mdcom_poll_rpu(void)
{
    kthread_run = true;

    while( kthread_run )
    {
        if( *MDCOM_RPU_NOTIFY == true )
        {
            *MDCOM_RPU_NOTIFY = false;
            mdcom_irq_handler(0, NULL);
        }
        msleep(RPU_POLL_TIME);
    }
}
#endif  /* USE_RPU_NOTIFY_POLLING */


/*
 *  mdcom_ipi_probe()
 *
 *  Function : This function initilizes this driver.
 *  Argument : pdev - platform device structer
 *  Return :   0 -     Success
 *             Other - Failed
 */
static int  mdcom_ipi_probe(struct platform_device *pdev)
{
    struct resource *res;
#ifndef USE_RPU_NOTIFY_POLLING
    int ret = 0;
#endif  /* USE_RPU_NOTIFY_POLLING */

    /*
     *  Assign IPI register
     */
    /* Get IPI register address from device tree */
    res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
    if (!res)  {
         pr_info("MDCOM: Failed - MDCOM Inter-processor interrupt Driver couldn't find resources.\n");
        return -ENODEV;
    }

    /* Assign IPI register */
    apu_ipi_base = devm_ioremap(&pdev->dev, res->start, resource_size(res));
    if( IS_ERR(apu_ipi_base) )
    {
        pr_err("%s: Unable to map IPI\n", __func__);
        return PTR_ERR(apu_ipi_base);
    }

	rpu0_ipi_base = apu_ipi_base + 0x00010000;

    pr_info("MDCOM: MDCOM Inter-processor interrupt Driver mapped IPI registers.\n");

#ifdef USE_RPU_NOTIFY_POLLING
	/* Create RPU polling thread */
	kthread = kthread_create(mdcom_poll_rpu, NULL, "mdcom_thread");
	if( IS_ERR(kthread) )
    {
        dev_err(&pdev->dev, "Failed to create kernel thread\n");
        return -ENODEV;
    }

	/* Start RPU polling thread */
	wake_up_process(kthread);
#else	/* USE_RPU_NOTIFY_POLLING */
    /*
     *  Initialize APU IPI
     */
    /* Disable RPU IPI interrupt */
    reg_write(apu_ipi_base, IDR_OFFSET, RPU_0_IPI_MASK);
    /* Clear RPU IPI interrupt */
    reg_write(apu_ipi_base, ISR_OFFSET, RPU_0_IPI_MASK);
    /* Enable RPU IPI interrupt */
    reg_write(apu_ipi_base, IER_OFFSET, RPU_0_IPI_MASK);

    pr_info("MDCOM: MDCOM Inter-processor interrupt Driver initialized APU IPI.\n");

    /*
     *  Register IPI IRQ
     */
    /* Get IPI IRQ  from device tree */
    ipi_irq = platform_get_irq(pdev, 0);
    if( ipi_irq < 0 )
    {
        dev_err(&pdev->dev, "unable to find IPI IRQ\n");
        iounmap(apu_ipi_base);
        apu_ipi_base = NULL;
        return ipi_irq;
    }

    /* Register MDCOM IRQ handler */
    ret = devm_request_irq(&pdev->dev, ipi_irq,
        mdcom_irq_handler, IRQF_SHARED, dev_name(&pdev->dev),
        &pdev->dev);
    if( ret )
    {
        dev_err(&pdev->dev, "IRQ %d already allocated\n", ipi_irq);
        iounmap(apu_ipi_base);
        apu_ipi_base = NULL;
        return ret;
    }

    pr_info("MDCOM: MDCOM Inter-processor interrupt Driver registered MDCOM IRQ handler.\n");
#endif  /* USE_RPU_NOTIFY_POLLING */

    return 0;
}


/*
 *  mdcom_ipi_remove()
 *
 *  Function : This function removes this driver.
 *  Argument : pdev - platform device structer
 *  Return :   0 -     Success
 *             Other - Failed
 */
static int  mdcom_ipi_remove(struct platform_device *pdev)
{
    iounmap(apu_ipi_base);
    apu_ipi_base = NULL;

#ifdef USE_RPU_NOTIFY_POLLING
    kthread_run = false;
    msleep(RPU_POLL_TIME + 10);
    kthread_stop(kthread);
    kthread = NULL;
#else   /* USE_RPU_NOTIFY_POLLING */
    devm_free_irq(&pdev->dev, ipi_irq, &pdev->dev);
    ipi_irq = 0xFFFFFFFF;
#endif  /* USE_RPU_NOTIFY_POLLING */

    return 0;
}


/*
 *  mdcom_ipi_generate_int()
 *
 *  Function : This function generates interupt to the master processor.
 *  Argument : None
 *  Return :   None
 */
void        mdcom_ipi_generate_int(void)
{
	volatile uint32_t	reg32;

	do
	{
	    wmb();
		reg32 = reg_read(apu_ipi_base, OBS_OFFSET);
	}
	while( reg32 & RPU_0_IPI_MASK );

    reg_write(apu_ipi_base, TRIG_OFFSET, RPU_0_IPI_MASK);
}
EXPORT_SYMBOL(mdcom_ipi_generate_int);


/*
 *  mdcom_ipi_clear_int()
 *
 *  Function : This function clears event of interupt.
 *  Argument : None
 *  Return :   None
 */
void        mdcom_ipi_clear_int(void)
{
#ifndef USE_RPU_NOTIFY_POLLING
    reg_write(apu_ipi_base, ISR_OFFSET, RPU_0_IPI_MASK);
	reg_write(rpu0_ipi_base, ISR_OFFSET, APU_IPI_MASK);
#endif  /* USE_RPU_NOTIFY_POLLING */
}
EXPORT_SYMBOL(mdcom_ipi_clear_int);


/*
 *  Tables
 */
static const struct of_device_id mdcom_ipi_match[] = {
    { .compatible = "xlnx,zynqmp-apu-ipi", },
    { /* end of list */ },
};
MODULE_DEVICE_TABLE(of, mdcom_ipi_match);

static struct platform_driver mdcom_ipi_driver = {
    .probe =  mdcom_ipi_probe,
    .remove = mdcom_ipi_remove,
    .driver = {
        .name = "mdcom_ipi",
        .of_match_table = mdcom_ipi_match,
    },
};
module_platform_driver(mdcom_ipi_driver);

MODULE_DESCRIPTION("MDCOM IPI driver");
MODULE_LICENSE("GPL v2");
